#include "test.hpp"

typedef TestTmpFile<docra::Lexer_CLike> TestCLikeTmpFile;

TEST_F(TestCLikeTmpFile, CLike0) {
  wl("//a test");
  wl("");
  wl("that's all");
  we();

  CASSERT_CODE("", 0);
  CASSERT_COMM("a test\n", 0);
  CASSERT_CODE("\n", 0);
  CASSERT_CODE("that's all\n", 0);
  CASSERT_EOF();
}

static void helloworld_body(TestCLikeTmpFile &l) {
  l.wl("//? This would be a longer description of what this function does");
  l.wl("//? It can extend multiple lines of course, and can mix code blocks, like this one: <?");
  l.wl("  if (argc > 1) { //? If the user passed an argument");
  l.wl("    printf(\"Hi %s! I'm %s\", argv[1], argv[0]);");
  l.wl("  } else { //? Otherwise ");
  l.wl("    printf(\"Hi! I'm %s\", argv[0]);");
  l.wl("  } //?>");
  l.wl("//? Code blocks and text can obviously be mixed.");
}

TEST_F(TestCLikeTmpFile, CLikeHello0) {
  wl("//+`main` Entry point of the executable");
  wl("int main(int argc, char *argv[]) {");
  wl("//-`argc` Number of arguments passed by the operating system");
  wl("//-`argv` Array of strings containing all arguments");
  helloworld_body(*this);
  ws("}");
  we();

  TASSERT_CODE(nullptr);
  TASSERT_TAG("+", "main", "");
  TASSERT_STR(nullptr);
  TASSERT_EOL();
  TASSERT_CODE(nullptr);
  TASSERT_EOL();
  TASSERT_SIN();
  TASSERT_CODE(nullptr);
  TASSERT_TAG("-", "argc", "");
  TASSERT_STR(nullptr);
  TASSERT_EOL();
  TASSERT_CODE(nullptr);
  TASSERT_TAG("-", "argv", "");
  TASSERT_STR(nullptr);
  TASSERT_EOL();
}

TEST_F(TestCLikeTmpFile, CLikeHello1) {
  wl("//+`main` Entry point of the executable");
  wl("int main(int argc, char *argv[]) {");
  wl("//-`argc` Number of arguments passed by the operating system");
  wl("//-`argv` Array of strings containing all arguments");
  helloworld_body(*this);
  ws("}");
  we();
}

TEST_F(TestCLikeTmpFile, CLikeHello2) {
  wl("//+`main` Entry point of the executable");
  wl("//-`argc` Number of arguments passed by the operating system");
  wl("//-`argv` Array of strings containing all arguments");
  wl("int main(int argc, char *argv[]) {");
  helloworld_body(*this);
  wl("}");
}

TEST_F(TestCLikeTmpFile, CLikeHello3) {
  wl("int main(int argc, char *argv[]) {");
  wl("//+`main` Entry point of the executable");
  wl("//-`argc` Number of arguments passed by the operating system");
  wl("//-`argv` Array of strings containing all arguments");
  helloworld_body(*this);
  wl("}");
}

TEST_F(TestCLikeTmpFile, CLikeHello4) {
  wl("int main(int argc, char *argv[]) { //+`main` Entry point of the executable");
  wl("//-`argc` Number of arguments passed by the operating system");
  wl("//-`argv` Array of strings containing all arguments");
  helloworld_body(*this);
  wl("}");
}

TEST_F(TestCLikeTmpFile, CLikeHello5) {
  wl("int main(int argc, char *argv[]) { //+`main`");
  wl("//? Entry point of the executable");
  wl("//-`argc` Number of arguments passed by the operating system");
  wl("//-`argv` Array of strings containing all arguments");
  helloworld_body(*this);
  wl("}");
}

static void complex_body(TestCLikeTmpFile &l) {
  l.wl("  double i; //+`.i` Imaginary part of this complex number");
  l.wl("  double r; //+`.r` Real part of this complex number");
}

TEST_F(TestCLikeTmpFile, ClikeLexComplex0) {
  wl("class Complex {//+`Complex`");
  wl("//? This is the short description");
  wl("//? This line, and the rest, will make up the long form description of this symbol.");
  complex_body(*this);
  wl("};");
  we();
}

TEST_F(TestCLikeTmpFile, ClikeLexComplex1) {
  wl("class Complex {");
  wl("//+`Complex`");
  wl("//? This is the short description");
  wl("//? This line, and the rest, will make up the long form description of this symbol.");
  complex_body(*this);
  wl("};");
  we();
}

TEST_F(TestCLikeTmpFile, ClikeLexComplex2) {
  wl("//+`Complex`");
  wl("//? This is the short description");
  wl("//? This line, and the rest, will make up the long form description of this symbol.");
  wl("class Complex {");
  complex_body(*this);
  wl("};");
  we();
}

static void CLikeParse_assert(docra::Project &p) {
  ASSERT_TRUE(p.locate("", "::test::A"));
  ASSERT_TRUE(p.locate("", "::test::B"));
  ASSERT_TRUE(p.locate("", "::test::B::C"));
  ASSERT_TRUE(p.locate("", "::test::B::D"));
}

TEST_F(TestCLikeTmpFile, CLikeParse0) {
  /* 1*/wl("namespace test { //`::test`+");
  /* 2*/wl("Bar X;");
  /* 3*/wl("");
  /* 4*/wl("Foo A; //+`::A` 1.");
  /* 5*/wl("Bar b; //+`::b` blabla");
  /* 6*/wl("");
  /* 7*/wl("struct B {");
  /* 8*/wl("  //+`::B` 5.");
  /* 9*/wl("  //this `should` be ignored!");
  /*10*/wl("  struct C {");
  /*11*/wl("    //+`::C` 5.");
  /*12*/wl("  };");
  /*13*/wl("  void D(E e); //+`::D` 1.");
  /*14*/wl("  //-`e` arg");
  /*15*/wl("};");
  /*16*/wl("} //namespace test");
  /*17*/we();

  docra::Scope root_scope;
  parser.parse(docra::Config::VERB_DEBUG, "", root_scope);

  CLikeParse_assert(parser.project);

  parser.project.debug();
}

TEST_F(TestCLikeTmpFile, CLikeParse1) {
  /* 1*/wl("namespace test { //`::test`+");
  /* 2*/wl("Bar X;");
  /* 3*/wl("");
  /* 4*/wl("Foo A;");
  /* 5*/wl("//+`::A` 4.");
  /* 6*/wl("struct B {");
  /* 7*/wl("  //+`::B` 5.");
  /* 8*/wl("  //+`::C` X.");
  /* 9*/wl("  struct C {");
  /*10*/wl("    //+`::C1` 5.");
  /*11*/wl("    typedef C* C1;");
  /*12*/wl("  };");
  /*13*/wl("  //+`::D` 1.");
  /*14*/wl("  void D(E e);");
  /*15*/wl("  //-`e` arg");
  /*16*/wl("};");
  /*17*/wl("} //namespace test");
  /*18*/we();

  docra::Scope root_scope;
  parser.parse(docra::Config::VERB_DEBUG, "", root_scope);

  CLikeParse_assert(parser.project);
  ASSERT_TRUE(parser.project.locate("", "::test::B::C::C1"));

  parser.project.debug();
}

TEST_F(TestCLikeTmpFile, CLikeParse2) {
  /* 1*/wl("namespace test { //`::test`+");
  /* 2*/wl("Bar X;");
  /* 3*/wl("");
  /* 4*/wl("//+`::A`");
  /* 5*/wl("Foo A;");
  /* 6*/wl("");
  /* 7*/wl("//+`::B`");
  /* 8*/wl("struct B {");
  /* 9*/wl("  struct C {");
  /*10*/wl("    //+`::C`");
  /*11*/wl("  };");
  /*12*/wl("  void D(E e); //+`::D`");
  /*13*/wl("  //-`e` arg");
  /*14*/wl("};");
  /*15*/wl("} //namespace test");
  /*16*/we();

  docra::Scope root_scope;
  parser.parse(docra::Config::VERB_DEBUG, "", root_scope);

  CLikeParse_assert(parser.project);

  parser.project.debug();
}

TEST_F(TestCLikeTmpFile, CLikeParse3) {
  /* 1*/wl("namespace test { //`::test`+");
  /* 2*/wl("Bar X;");
  /* 3*/wl("");
  /* 4*/wl("Foo A; //+`::A`");
  /* 5*/wl("");
  /* 6*/wl("//+`::B`");
  /* 7*/wl("struct B {");
  /* 8*/wl("  //+`::C`");
  /* 9*/wl("  struct C {");
  /*10*/wl("  };");
  /*11*/wl("  void D(E e); //+`::D`");
  /*12*/wl("  //-`e` arg");
  /*13*/wl("};");
  /*14*/wl("} //namespace test");
  /*15*/we();

  docra::Scope root_scope;
  parser.parse(docra::Config::VERB_DEBUG, "", root_scope);

  CLikeParse_assert(parser.project);

  parser.project.debug();
}

TEST_F(TestCLikeTmpFile, CLikeParse4) {
  /* 1*/wl("namespace test { //`::test`+");
  /* 2*/wl("Bar X;");
  /* 3*/wl("");
  /* 4*/wl("Foo A; //+`::A`");
  /* 5*/wl("");
  /* 6*/wl("struct B { //+`::B`");
  /* 7*/wl("  //+`::C`");
  /* 8*/wl("  struct C {");
  /* 9*/wl("  };");
  /*10*/wl("  void D(E e); //+`::D`");
  /*11*/wl("  //-`e` arg");
  /*12*/wl("};");
  /*13*/wl("} //namespace test");
  /*14*/we();

  docra::Scope root_scope;
  parser.parse(docra::Config::VERB_DEBUG, "", root_scope);

  CLikeParse_assert(parser.project);

  parser.project.debug();
}

TEST_F(TestCLikeTmpFile, CLikeParse5) {
  /* 1*/wl("//`test` This is a tag");
  /* 2*/wl("//? Followed by a comment");
  /* 3*/wl("int test = 0;");
  /* 4*/we();

  docra::Scope root_scope;
  parser.parse(docra::Config::VERB_DEBUG, "", root_scope);

  parser.project.debug();

  ASSERT_TRUE(parser.project.locate("", "test"));
}

TEST_F(TestCLikeTmpFile, CLikeParse6) {
  /* 1*/wl("//`prefix`+");
  /* 2*/wl("");
  /* 3*/wl("void prefix::foo() {");
  /* 4*/wl("  // an empty scope");
  /* 5*/wl("}");
  /* 6*/wl("");
  /* 7*/wl("//+`::bar()`");
  /* 8*/wl("//>`someinput`");
  /* 9*/wl("//-> return type");
  /*10*/wl("void prefix::bar() {");
  /*11*/wl("");
  /*12*/wl("}");
  we();

  docra::Scope root_scope;
  parser.parse(docra::Config::VERB_DEBUG, "", root_scope);

  parser.project.debug();

  docra::Suffix *bar = parser.project.locate("", "prefix::bar()");
  ASSERT_TRUE(bar);
  ASSERT_TRUE(bar->lineno_code > bar->lineno_from());
}

TEST_F(TestCLikeTmpFile, CLikeParse7) {

  /* 1*/wl("//+`::Token`");
  /* 2*/wl("struct Token {");
  /* 3*/wl("  enum KIND {//+`::KIND` The specific kind of `Token`");
  /* 4*/wl("    CODE,        //+`.CODE` A single line of code");
  /* 5*/wl("    // A `Chunk` of `Chunk.CODE` kind goes through the `Lexer` without modification");
  /* 6*/wl("    SCOPE_ENTER, //+`.SCOPE_ENTER` Emitted when entering a scope");
  /* 7*/wl("    // See #`scoping`");
  /* 8*/wl("    EOL          //+`.EOL` End-of-line");
  /* 9*/wl("  };");
  /*10*/wl("};");
  we();

  docra::Scope root_scope;
  parser.parse(docra::Config::VERB_DEBUG, "", root_scope);

  parser.project.debug();
}
